/*************************************************************************/
/* File:        filter.v
 * Author:      Carl Ebeling
 *
 * Description: This is a template for the 2D wavelet filter for our camera
 * Revsions:
 *   1.0        Initial design - 1/27/2009 - Carl Ebeling
 **************************************************************************/
`timescale 1ns/1ns

module filter
  (input clk,
   input enable,
   input newFrame,
   input pixelValidIn,
   input [7:0] lineAddrIn,
   input [8:0] colAddrIn,
   input [7:0] pixelRedIn,
   input [7:0] pixelGreenIn,
   input [7:0] pixelBlueIn,
   output reg pixelValidOut,
   output reg [7:0] lineAddrOut,
   output reg [8:0] colAddrOut,
   output reg [7:0] pixelRedOut,
   output reg [7:0] pixelGreenOut,
   output reg [7:0] pixelBlueOut
  );

   wire [7:0] 	gray = (307*pixelRedIn + 604*pixelGreenIn + 113*pixelBlueIn + 512) >> 10 ;
   wire [7:0] 	evenGray;   // Pixel from the even row
   
   reg [7:0] 	grayOut;    // Pixel value for output
   reg [8:0] 	colAddr,    // Column address for output
				colAddrSave;   // Save because it will change while we are doing output
   reg [7:0] 	lineAddr,   // Line address for output
				lineAddrSave; // Save because it will change while we are doing output
   
   // These registers are used to store the common subexpressions
   reg signed [9:0] 	D00pD10t, D00pD10;
   reg signed [9:0] 	D00mD10t, D00mD10;
   reg signed [9:0] 	D01pD11, D01mD11;
   // State machine register
   reg [2:0] 	state;
   localparam IDLE=0, Q00state=1, Q01state=2, Q10state=3, Q11state=4;
   
   // Use this memory to save the pixels in the even rows
   // + Write valid pixels to this memory only when the row is even
   // + During odd rows, the pixels read out will be from the even row above
   // + (During even rows, pixels read will not be used)
   rowMem row
	 (.clk (clk),
	  .write (/* FILL IN */  ),
	  .writeAddr (/* FILL IN */), 
	  .writeData (gray),    // Even row pixels written in
	  .readAddr (/* FILL IN */), 
	  .readData (evenGray)); // Even row pixels read out

   always @(posedge clk) begin
	  // Default
	  state <= IDLE;
	  // State machine - rotate through the states
	  case (state) 
		Q00state: state <= Q01state;
		Q01state: state <= Q10state;
		Q10state: state <= Q11state;
		Q11state: state <= IDLE;
	  endcase // case (state)

	  if (lineAddrIn[0]) begin
		 if (pixelValidIn) begin
  			if (colAddrIn[0] == 0) begin
			   // First column of 2x2 block: compute 1st set common subexpressions
			   D00pD10t <= /* FILL IN */ 0;
			   D00mD10t <= /* FILL IN */ 0;
			end else begin
			   // Second column of 2x2 block: compute 2nd set of common subexpressions
			   D00pD10 <= D00pD10t;          // Must save so it doesn't get clobbered??
			   D00mD10 <= D00mD10t;
			   D01pD11 <= /* FILL IN */ 0;
			   D01mD11 <= /* FILL IN */ 0;
			   // Save the column address of this block
			   colAddrSave <= colAddrIn; 
			   lineAddrSave <= lineAddrIn;
			   // Start the state machine to write the Q outputs to the frame buffer
			   if (state == IDLE || state == Q11state) state <= Q00state;  // We should be done with the previous Qs!
			end
		 end
	  end
   end
   // Depending on which state we are in, generate the output Qs
   always @(*) begin
	  case (state) 
		Q00state: begin
		   /* FILL IN */
		end
		Q01state: begin
		   /* FILL IN */
		end
		Q10state: begin
		   /* FILL IN */
		end
		Q11state: begin
		   /* FILL IN */
		end
		default: begin
		   grayOut = 0;
		   lineAddr = 0;
		   colAddr = 0;
		end
	  endcase // case (state)

	  if (enable) begin
		 pixelRedOut = grayOut;
		 pixelGreenOut = grayOut;
		 pixelBlueOut = grayOut;
		 lineAddrOut = lineAddr;
		 colAddrOut = colAddr;
		 pixelValidOut = (state != IDLE);
	  end else begin
// Pass through
		 pixelValidOut = pixelValidIn;
		 lineAddrOut = lineAddrIn;
		 colAddrOut = colAddrIn;
// Use this for simple pass through	of color image
		 pixelRedOut = pixelRedIn;
		 pixelGreenOut = pixelGreenIn;
		 pixelBlueOut = pixelBlueIn;
// Use this for gray scale image
//		 pixelRedOut = gray;
//		 pixelGreenOut = gray;
//		 pixelBlueOut = gray;
// Use this code to test that memory is correct
//		 pixelRedOut = evenGray; 
//		 pixelGreenOut = evenGray;
//		 pixelBlueOut = evenGray;
	  end
   end // always @ (*)

endmodule
